@@ -13,6 +13,11 @@
 #include "printtags.h"
 #include "routines.h"
 #include "routines_p.h"
+
+#include "vstring.h"
+#include "htable.h"
+#include "fname.h"
+
 #include <string.h>		/* strerror */
 #include <stdlib.h>		/* exit */
 #include <stdio.h>		/* stderr */
@@ -24,6 +29,12 @@ typedef struct sReadOption {
 	int matchOpts;
 } readOptions;
 
+struct canonWorkArea {
+	struct canonFnameCacheTable *cacheTable;
+	int ptags;
+	bool absoluteOnly;
+};
+
 static const char *TagFileName = "tags";
 static const char *ProgramName;
 static int debugMode;
@@ -141,7 +152,7 @@ struct tagEntryArray {
 	struct tagEntryHolder *a;
 };
 
-struct tagEntryArray *tagEntryArrayNew (void)
+static struct tagEntryArray *tagEntryArrayNew (void)
 {
 	struct tagEntryArray * a = eMalloc (sizeof (struct tagEntryArray));
 
@@ -152,7 +163,7 @@ struct tagEntryArray *tagEntryArrayNew (void)
 	return a;
 }
 
-void tagEntryArrayPush (struct tagEntryArray *a, tagEntry *e)
+static void tagEntryArrayPush (struct tagEntryArray *a, tagEntry *e)
 {
 	if (a->count + 1 == a->length)
 	{
@@ -169,7 +180,7 @@ void tagEntryArrayPush (struct tagEntryArray *a, tagEntry *e)
 	a->a[a->count++].e = e;
 }
 
-void tagEntryArrayFree (struct tagEntryArray *a, int freeTags)
+static void tagEntryArrayFree (struct tagEntryArray *a, int freeTags)
 {
 	if (freeTags)
 	{
@@ -187,7 +198,8 @@ static int compareTagEntry (const void *a, const void *b)
 
 static void walkTags (tagFile *const file, tagEntry *first_entry,
 					  tagResult (* nextfn) (tagFile *const, tagEntry *),
-					  void (* actionfn) (const tagEntry *, void *), void *data)
+					  void (* actionfn) (const tagEntry *, void *), void *data,
+					  struct canonWorkArea *canon)
 {
 	struct tagEntryArray *a = NULL;
 
@@ -196,9 +208,21 @@ static void walkTags (tagFile *const file, tagEntry *first_entry,
 
 	do
 	{
+		tagEntry *shadow = first_entry;
+		tagEntry  shadowRec;
+		if (canon
+			&& (canon->ptags == 0
+				|| strcmp (first_entry->name, "!_TAG_PROC_CWD") == 0))
+		{
+			shadowRec = *first_entry;
+			shadow = &shadowRec;
+			shadow->file = canonicalizeFileName (canon->cacheTable,
+												 first_entry->file);
+		}
+
 		if (Qualifier)
 		{
-			int i = q_is_acceptable (Qualifier, first_entry);
+			int i = q_is_acceptable (Qualifier, shadow);
 			switch (i)
 			{
 			case Q_REJECT:
@@ -210,17 +234,17 @@ static void walkTags (tagFile *const file, tagEntry *first_entry,
 
 		if (a)
 		{
-			tagEntry *e = copyTag (first_entry);
+			tagEntry *e = copyTag (shadow);
 			tagEntryArrayPush (a, e);
 		}
 		else
-			(* actionfn) (first_entry, data);
+			(* actionfn) (shadow, data);
 	} while ( (*nextfn) (file, first_entry) == TagSuccess);
 
 	int err = tagsGetErrno (file);
 	if (err != 0)
 	{
-		fprintf (stderr, "%s: error in walktTags(): %s\n",
+		fprintf (stderr, "%s: error in walkTags(): %s\n",
 				 ProgramName,
 				 tagsStrerror (err));
 		exit (1);
@@ -237,16 +261,31 @@ static void walkTags (tagFile *const file, tagEntry *first_entry,
 #else
 static void walkTags (tagFile *const file, tagEntry *first_entry,
 					  tagResult (* nextfn) (tagFile *const, tagEntry *),
-					  void (* actionfn) (const tagEntry *, void *), void *data)
+					  void (* actionfn) (const tagEntry *, void *), void *data,
+					  struct canonWorkArea *canon)
 {
 	do
-		(* actionfn) (first_entry, data);
+	{
+		tagEntry *shadow = first_entry;
+		tagEntry  shadowRec;
+		if (canon
+			&& (canon->ptags == 0
+				|| strcmp (first_entry->name, "!_TAG_PROC_CWD") == 0))
+		{
+			shadow = &shadowRec;
+			shadowRec = *first_entry;
+			shadow->file = canonicalizeFileName (canon->cacheTable,
+												 first_entry->file);
+		}
+
+		(* actionfn) (shadow, data);
+	}
 	while ( (*nextfn) (file, first_entry) == TagSuccess);
 
 	int err = tagsGetErrno (file);
 	if (err != 0)
 	{
-		fprintf (stderr, "%s: error in walktTags(): %s\n",
+		fprintf (stderr, "%s: error in walkTags(): %s\n",
 				 ProgramName,
 				 tagsStrerror (err));
 		exit (1);
@@ -290,6 +329,43 @@ static void removeTagFile (void)
 	eFree ((char *)TagFileName);
 }
 
+static const char *loadCtagsCWD (tagFile *const file, tagEntry *pentry)
+{
+	if (tagsFindPseudoTag (file, pentry, "!_TAG_PROC_CWD",
+						   TAG_FULLMATCH) != TagSuccess)
+	{
+		int err = tagsGetErrno (file);
+		if (!err)
+		{
+			fprintf (stderr, "%s: no !_TAG_PROC_CWD in %s\n",
+					 ProgramName, TagFileName);
+			exit (1);
+		}
+
+		fprintf (stderr, "%s: cannot find !_TAG_PROC_CWD in %s: %s\n",
+				 ProgramName, TagFileName, tagsStrerror (err));
+		exit (1);
+	}
+
+	if (pentry->file[0] != '/')
+	{
+		fputs ("!_TAG_PROC_CWD must start with '/': ", stderr);
+		tagsPrintValue (pentry->file, 1, NULL, stderr);
+		fputc ('\n', stderr);
+		exit (1);
+	}
+
+	return pentry->file;
+}
+
+static struct canonFnameCacheTable *makeCanonFnameCacheTable (tagFile *const file,
+															  bool absoluteOnly)
+{
+	tagEntry pentry;
+	const char *cwd = loadCtagsCWD (file, &pentry);
+	return canonFnameCacheTableNew (cwd, absoluteOnly);
+}
+
 static tagFile *openTags (const char *const filePath, tagFileInfo *const info)
 {
 	if (strcmp (filePath, "-") == 0)
@@ -336,7 +412,7 @@ static int hasPsuedoTag (tagFile *const file,
 }
 
 static void findTag (const char *const name, readOptions *readOpts,
-					 tagPrintOptions *printOpts)
+					 tagPrintOptions *printOpts, struct canonWorkArea *canon)
 {
 	tagFileInfo info;
 	tagEntry entry;
@@ -351,6 +427,9 @@ static void findTag (const char *const name, readOptions *readOpts,
 		exit (1);
 	}
 
+	if (canon && canon->cacheTable == NULL)
+		canon->cacheTable = makeCanonFnameCacheTable (file, canon->absoluteOnly);
+
 	if (printOpts->escaping)
 	{
 		printOpts->escapingInputField = 0;
@@ -379,7 +458,8 @@ static void findTag (const char *const name, readOptions *readOpts,
 #ifdef READTAGS_DSL
 				  Formatter? printTagWithFormatter:
 #endif
-				  printTag, printOpts);
+				  printTag, printOpts,
+				  canon);
 	else if ((err = tagsGetErrno (file)) != 0)
 	{
 		fprintf (stderr, "%s: error in tagsFind(): %s\n",
@@ -390,7 +470,8 @@ static void findTag (const char *const name, readOptions *readOpts,
 	tagsClose (file);
 }
 
-static void listTags (int pseudoTags, tagPrintOptions *printOpts)
+static void listTags (int pseudoTags, tagPrintOptions *printOpts,
+					  struct canonWorkArea *canon)
 {
 	tagFileInfo info;
 	tagEntry entry;
@@ -407,6 +488,9 @@ static void listTags (int pseudoTags, tagPrintOptions *printOpts)
 		exit (1);
 	}
 
+	if (canon && canon->cacheTable == NULL)
+		canon->cacheTable = makeCanonFnameCacheTable (file, canon->absoluteOnly);
+
 	if (printOpts->escaping)
 	{
 		printOpts->escapingInputField = 0;
@@ -418,8 +502,8 @@ static void listTags (int pseudoTags, tagPrintOptions *printOpts)
 	if (pseudoTags)
 	{
 		if (tagsFirstPseudoTag (file, &entry) == TagSuccess)
-			walkTags (file, &entry, tagsNextPseudoTag, printPseudoTag,
-					  printOpts);
+			walkTags (file, &entry, tagsNextPseudoTag, printPseudoTag, printOpts,
+					  canon);
 		else if ((err = tagsGetErrno (file)) != 0)
 		{
 			fprintf (stderr, "%s: error in tagsFirstPseudoTag(): %s\n",
@@ -435,7 +519,8 @@ static void listTags (int pseudoTags, tagPrintOptions *printOpts)
 #ifdef READTAGS_DSL
 					  Formatter? printTagWithFormatter:
 #endif
-					  printTag, printOpts);
+					  printTag, printOpts,
+					  canon);
 		else if ((err = tagsGetErrno (file)) != 0)
 		{
 			fprintf (stderr, "%s: error in tagsFirst(): %s\n",
@@ -482,12 +567,18 @@ static const char *const Usage =
 	"        Also include the line number field when -e option is given.\n"
 	"    -p | --prefix-match\n"
 	"        Perform prefix matching in the NAME action.\n"
+	"    -P | --with-pseudo-tags\n"
+	"        List pseudo tags as if -D option is specified but continues processing without exiting.\n"
 	"    -t TAGFILE | --tag-file TAGFILE\n"
 	"        Use specified tag file (default: \"tags\").\n"
 	"        \"-\" indicates taking tag file data from standard input.\n"
 	"    -s[0|1|2] | --override-sort-detection METHOD\n"
 	"        Override sort detection of tag file.\n"
 	"        METHOD: unsorted|sorted|foldcase\n"
+	"    -C | --canonicalize-input\n"
+	"        Reduct '..' and '.' in input fields.\n"
+	"    -A | --absolute-input\n"
+	"        Do the same as -C but use absolute path form\n"
 #ifdef READTAGS_DSL
 	"    -F EXP | --formatter EXP\n"
 	"        Format the tags listed by ACTION with EXP when printing.\n"
@@ -571,12 +662,26 @@ extern int main (int argc, char **argv)
 	int actionSupplied = 0;
 	int i;
 	int ignore_prefix = 0;
+
 	tagPrintOptions printOpts;
 	readOptions readOpts;
 
 	memset (&printOpts, 0, sizeof (printOpts));
 	memset (&readOpts, 0, sizeof (readOpts));
 
+	struct canonWorkArea canonWorkArea = {
+		.cacheTable = NULL,
+		.ptags = 0,
+		.absoluteOnly = false,
+	};
+	struct canonWorkArea canonWorkAreaAbsForm = {
+		.cacheTable = NULL,
+		.ptags = 0,
+		.absoluteOnly = true,
+	};
+	struct canonWorkArea *canon = NULL;
+
+
 	ProgramName = argv [0];
 	setExecutableName (ProgramName);
 	if (argc == 1)
@@ -586,7 +691,9 @@ extern int main (int argc, char **argv)
 		const char *const arg = argv [i];
 		if (ignore_prefix || arg [0] != '-')
 		{
-			findTag (arg, &readOpts, &printOpts);
+			if (canon)
+				canon->ptags = 0;
+			findTag (arg, &readOpts, &printOpts, canon);
 			actionSupplied = 1;
 		}
 		else if (arg [0] == '-' && arg [1] == '\0')
@@ -596,10 +703,14 @@ extern int main (int argc, char **argv)
 			const char *optname = arg + 2;
 			if (strcmp (optname, "debug") == 0)
 				debugMode++;
-			else if (strcmp (optname, "list-pseudo-tags") == 0)
+			else if (strcmp (optname, "list-pseudo-tags") == 0
+					 || strcmp (optname, "with-pseudo-tags") == 0)
 			{
-				listTags (1, &printOpts);
-				actionSupplied = 1;
+				if (canon)
+					canon->ptags = 1;
+				listTags (1, &printOpts, NULL);
+				if (optname[0] == 'l')
+					actionSupplied = 1;
 			}
 			else if (strcmp (optname, "help") == 0)
 				printUsage (stdout, 0);
@@ -643,7 +754,9 @@ extern int main (int argc, char **argv)
 				readOpts.matchOpts |= TAG_PARTIALMATCH;
 			else if (strcmp (optname, "list") == 0)
 			{
-				listTags (0, &printOpts);
+				if (canon)
+					canon->ptags = 0;
+				listTags (0, &printOpts, canon);
 				actionSupplied = 1;
 			}
 			else if (strcmp (optname, "line-number") == 0)
@@ -683,6 +796,10 @@ extern int main (int argc, char **argv)
 					exit (1);
 				}
 			}
+			else if (strcmp (optname, "absolute-input") == 0)
+				canon = &canonWorkAreaAbsForm;
+			else if (strcmp (optname, "canonicalize-input") == 0)
+				canon = &canonWorkArea;
 #ifdef READTAGS_DSL
 			else if (strcmp (optname, "filter") == 0)
 			{
@@ -713,9 +830,9 @@ extern int main (int argc, char **argv)
 			else if (strcmp (optname, "formatter") == 0)
 			{
 				if (i + 1 < argc)
-					Sorter = compileExpression (argv[++i],
-												(void * (*)(EsObject *))f_compile,
-												optname);
+					Formatter = compileExpression (argv[++i],
+												   (void * (*)(EsObject *))f_compile,
+												   optname);
 				else
 				{
 					fprintf (stderr, "%s: missing formatter expression for --%s option\n",
@@ -740,7 +857,14 @@ extern int main (int argc, char **argv)
 				switch (arg [j])
 				{
 					case 'd': debugMode++; break;
-					case 'D': listTags (1, &printOpts); actionSupplied = 1; break;
+					case 'D':
+					case 'P':
+						if (canon)
+							canon->ptags = 1;
+						listTags (1, &printOpts, canon);
+						if (arg  [j] == 'D')
+							actionSupplied = 1;
+						break;
 					case 'h': printUsage (stdout, 0); break;
 #ifdef READTAGS_DSL
 					case 'H':
@@ -764,7 +888,12 @@ extern int main (int argc, char **argv)
 					case 'e': printOpts.extensionFields = 1; break;
 					case 'i': readOpts.matchOpts |= TAG_IGNORECASE;   break;
 					case 'p': readOpts.matchOpts |= TAG_PARTIALMATCH; break;
-					case 'l': listTags (0, &printOpts); actionSupplied = 1; break;
+					case 'l':
+						if (canon)
+							canon->ptags = 0;
+						listTags (0, &printOpts, canon);
+						actionSupplied = 1;
+						break;
 					case 'n': printOpts.lineNumber = 1; break;
 					case 't':
 						if (arg [j+1] != '\0')
@@ -787,6 +916,12 @@ extern int main (int argc, char **argv)
 						else
 							printUsage(stderr, 1);
 						break;
+					case 'A':
+						canon = &canonWorkAreaAbsForm;
+						break;
+					case 'C':
+						canon = &canonWorkArea;
+						break;
 #ifdef READTAGS_DSL
 					case 'Q':
 						if (i + 1 == argc)
@@ -834,5 +969,11 @@ extern int main (int argc, char **argv)
 	if (Formatter)
 		f_destroy (Formatter);
 #endif
+
+	if (canon)
+	{
+		if (canon->cacheTable)
+			canonFnameCacheTableDelete (canon->cacheTable);
+	}
 	return 0;
 }
