@@ -44,6 +44,7 @@
 #include "field.h"
 #include "fmt_p.h"
 #include "kind.h"
+#include "interval_tree_generic.h"
 #include "nestlevel.h"
 #include "options_p.h"
 #include "ptag_p.h"
@@ -74,7 +75,7 @@
 
 /*  Hack for ridiculous practice of Microsoft Visual C++.
  */
-#if defined (WIN32) && defined (_MSC_VER)
+#if defined (_WIN32) && defined (_MSC_VER)
 # define chsize         _chsize
 # define open           _open
 # define close          _close
@@ -95,6 +96,7 @@ typedef struct eTagFile {
 	int cork;
 	unsigned int corkFlags;
 	ptrArray *corkQueue;
+	struct rb_root intervaltab;
 
 	bool patternCacheValid;
 } tagFile;
@@ -104,6 +106,8 @@ typedef struct sTagEntryInfoX  {
 	int corkIndex;
 	struct rb_root symtab;
 	struct rb_node symnode;
+	struct rb_node intervalnode;
+	unsigned long __intervalnode_subtree_last;
 } tagEntryInfoX;
 
 /*
@@ -119,6 +123,13 @@ static tagFile TagFile = {
 	NULL,                /* vLine */
 	.cork = false,
 	.corkQueue = NULL,
+	/* .intervaltab = RB_ROOT,
+	 *
+	 * msvc doesn't accept the above expression:
+	 *
+	 *   main\entry.c(128) : error C2099: initializer is not a constant
+	 *
+	 */
 	.patternCacheValid = false,
 };
 
@@ -135,6 +146,13 @@ extern int truncate (const char *path, off_t length);
 extern int ftruncate (int fd, off_t length);
 #endif
 
+#define INTERVAL_START(node) ((node)->slot.lineNumber)
+#define INTERVAL_END(node)   ((node)->slot.extensionFields._endLine)
+
+INTERVAL_TREE_DEFINE(tagEntryInfoX, intervalnode,
+					 unsigned long, __intervalnode_subtree_last,
+					 INTERVAL_START, INTERVAL_END, /*static*/, intervaltab)
+
 /*
 *   FUNCTION DEFINITIONS
 */
@@ -200,7 +218,7 @@ extern void makeFileTag (const char *const fileName)
 		   unnecessary read line loop. */
 		while (readLineFromInputFile () != NULL)
 			; /* Do nothing */
-		tag.extensionFields.endLine = getInputLineNumber ();
+		setTagEndLine(&tag, getInputLineNumber ());
 	}
 
 	if (isFieldEnabled (FIELD_EPOCH))
@@ -669,7 +687,7 @@ static size_t appendInputLine (int putc_func (char , void *), const char *const
 		const int next = *(p + 1);
 		const int c = *p;
 
-		if (c == CRETURN  ||  c == NEWLINE)
+		if (c == '\r'  ||  c == '\n')
 			break;
 
 		if (patternLengthLimit != 0 && length >= patternLengthLimit &&
@@ -683,10 +701,10 @@ static size_t appendInputLine (int putc_func (char , void *), const char *const
 		}
 		/*  If character is '\', or a terminal '$', then quote it.
 		 */
-		if (c == BACKSLASH  ||  c == (Option.backward ? '?' : '/')  ||
-			(c == '$'  &&  (next == NEWLINE  ||  next == CRETURN)))
+		if (c == '\\'  ||  c == (Option.backward ? '?' : '/')  ||
+			(c == '$'  &&  (next == '\n'  ||  next == '\r')))
 		{
-			putc_func (BACKSLASH, data);
+			putc_func ('\\', data);
 			++length;
 		}
 		putc_func (c, data);
@@ -1020,11 +1038,11 @@ extern const tagField* getParserFieldForIndex (const tagEntryInfo * tag, int ind
 	}
 }
 
-extern const char* getParserFieldValueForType (tagEntryInfo *const tag, fieldType ftype)
+extern const char* getParserFieldValueForType (const tagEntryInfo *const tag, fieldType ftype)
 {
-	for (int i = 0; i < tag->usedParserFields; i++)
+	for (unsigned int i = 0; i < tag->usedParserFields; i++)
 	{
-		const tagField *f = getParserFieldForIndex (tag, i);
+		const tagField *f = getParserFieldForIndex (tag, (int)i);
 		if (f && f->ftype == ftype)
 			return f->value;
 	}
@@ -1061,16 +1079,33 @@ static tagEntryInfo *newNilTagEntry (unsigned int corkFlags)
 	x->slot.kindIndex = KIND_FILE_INDEX;
 	x->slot.inputFileName = getInputFileName ();
 	x->slot.inputFileName = eStrdup (x->slot.inputFileName);
+	x->slot.sourceFileName = getSourceFileTagPath();
+	if (x->slot.sourceFileName)
+		x->slot.sourceFileName = eStrdup (x->slot.sourceFileName);
 	return &(x->slot);
 }
 
+static void copyExtraDynamic (const tagEntryInfo *const src, tagEntryInfo *const dst)
+{
+	if (dst->extraDynamic)
+	{
+		unsigned int n = countXtags () - XTAG_COUNT;
+		dst->extraDynamic = xCalloc ((n / 8) + 1, uint8_t);
+		memcpy (dst->extraDynamic, src->extraDynamic, (n / 8) + 1);
+		PARSER_TRASH_BOX(dst->extraDynamic, eFree);
+	}
+}
+
 static tagEntryInfoX *copyTagEntry (const tagEntryInfo *const tag,
 									const char *shareInputFileName,
+									const char *sharedSourceFileName,
 									unsigned int corkFlags)
 {
 	tagEntryInfoX *x = xMalloc (1, tagEntryInfoX);
 	x->symtab = RB_ROOT;
 	x->corkIndex = CORK_NIL;
+	memset(&x->intervalnode, 0, sizeof (x->intervalnode));
+	x->__intervalnode_subtree_last = 0;
 	tagEntryInfo  *slot = (tagEntryInfo *)x;
 
 	*slot = *tag;
@@ -1109,16 +1144,29 @@ static tagEntryInfoX *copyTagEntry (const tagEntryInfo *const tag,
 		slot->extensionFields.xpath = eStrdup (slot->extensionFields.xpath);
 #endif
 
+	copyExtraDynamic (tag, slot);
 	if (slot->extraDynamic)
+		PARSER_TRASH_BOX_TAKE_BACK(slot->extraDynamic);
+
+	if (slot->sourceFileName == NULL)
+		slot->isSourceFileNameShared = 0;
+	else if (strcmp(slot->sourceFileName, sharedSourceFileName) == 0)
 	{
-		int n = countXtags () - XTAG_COUNT;
-		slot->extraDynamic = xCalloc ((n / 8) + 1, uint8_t);
-		memcpy (slot->extraDynamic, tag->extraDynamic, (n / 8) + 1);
+		/* strcmp() is needed here.
+		 * sharedSourceFileName can be changed during parsing a file.
+		 * So we cannot use the condition like:
+		 *
+		 *    if (slot->sourceFileName == getSourceFileTagPath()) { ... }
+		 *
+		 */
+		slot->sourceFileName = sharedSourceFileName;
+		slot->isSourceFileNameShared = 1;
 	}
-
-	if (slot->sourceFileName)
+	else
+	{
 		slot->sourceFileName = eStrdup (slot->sourceFileName);
-
+		slot->isSourceFileNameShared = 0;
+	}
 
 	slot->usedParserFields = 0;
 	slot->parserFieldsDynamic = NULL;
@@ -1161,6 +1209,8 @@ static void deleteTagEnry (void *data)
 	if (slot->kindIndex == KIND_FILE_INDEX)
 	{
 		eFree ((char *)slot->inputFileName);
+		if (slot->sourceFileName)
+			eFree ((char *)slot->sourceFileName);
 		goto out;
 	}
 
@@ -1194,7 +1244,7 @@ static void deleteTagEnry (void *data)
 	if (slot->extraDynamic)
 		eFree (slot->extraDynamic);
 
-	if (slot->sourceFileName)
+	if (slot->sourceFileName && !slot->isSourceFileNameShared)
 		eFree ((char *)slot->sourceFileName);
 
 	clearParserFields (slot);
@@ -1535,7 +1585,7 @@ static int queueTagEntry (const tagEntryInfo *const tag)
 
 	int corkIndex;
 	tagEntryInfo * nil = ptrArrayItem (TagFile.corkQueue, 0);
-	tagEntryInfoX * entry = copyTagEntry (tag, nil->inputFileName,
+	tagEntryInfoX * entry = copyTagEntry (tag, nil->inputFileName, nil->sourceFileName,
 										TagFile.corkFlags);
 
 	if (ptrArrayCount (TagFile.corkQueue) == (size_t)INT_MAX)
@@ -1556,9 +1606,82 @@ static int queueTagEntry (const tagEntryInfo *const tag)
 	entry->corkIndex = corkIndex;
 	entry->slot.inCorkQueue = 1;
 
+	/* Don't put FQ tags to interval table.
+	 * About placeholder, a parser should call
+	 * removeFromIntervalTabMaybe() explicitly.
+	 */
+	if (! isTagExtraBitMarked (&entry->slot, XTAG_FILE_NAMES)
+		&& entry->slot.extensionFields._endLine > entry->slot.lineNumber
+		&& !isTagExtraBitMarked (tag, XTAG_QUALIFIED_TAGS))
+	{
+		intervaltab_insert(entry, &TagFile.intervaltab);
+		entry->slot.inIntevalTab = 1;
+	}
 	return corkIndex;
 }
 
+extern void updateTagLine (tagEntryInfo *tag, unsigned long lineNumber,
+						   MIOPos filePosition)
+{
+	tagEntryInfoX *entry = NULL;
+	if (tag->inIntevalTab)
+	{
+		entry = (tagEntryInfoX *)tag;
+		removeFromIntervalTabMaybe (entry->corkIndex);
+	}
+
+	tag->lineNumber = lineNumber;
+	tag->filePosition = filePosition;
+	tag->boundaryInfo = getNestedInputBoundaryInfo (lineNumber);
+
+	if (entry && tag->lineNumber < tag->extensionFields._endLine)
+	{
+		intervaltab_insert(entry, &TagFile.intervaltab);
+		tag->inIntevalTab = 1;
+	}
+}
+
+extern void setTagEndLine(tagEntryInfo *tag, unsigned long endLine)
+{
+	if (endLine != 0 && endLine < tag->lineNumber)
+	{
+		error (WARNING,
+			   "given end line (%lu) for the tag (%s) in the file (%s) is smaller than its start line: %lu",
+			   endLine,
+			   tag->name,
+			   tag->inputFileName,
+			   tag->lineNumber);
+		return;
+	}
+
+	if (isTagExtraBitMarked (tag, XTAG_FILE_NAMES)
+		|| !tag->inCorkQueue
+		|| isTagExtraBitMarked (tag, XTAG_QUALIFIED_TAGS))
+	{
+		tag->extensionFields._endLine = endLine;
+		return;
+	}
+
+	tagEntryInfoX *entry = (tagEntryInfoX *)tag;
+
+	if (tag->inIntevalTab)
+		removeFromIntervalTabMaybe (entry->corkIndex);
+
+	tag->extensionFields._endLine = endLine;
+	if (endLine > tag->lineNumber)
+	{
+		intervaltab_insert(entry, &TagFile.intervaltab);
+		tag->inIntevalTab = 1;
+	}
+}
+
+extern void setTagEndLineToCorkEntry (int corkIndex, unsigned long endLine)
+{
+	tagEntryInfo *entry = getEntryInCorkQueue (corkIndex);
+	if (entry)
+		setTagEndLine (entry, endLine);
+}
+
 extern void setupWriter (void *writerClientData)
 {
 	writerSetup (TagFile.mio, writerClientData);
@@ -1574,6 +1697,9 @@ static bool isTagWritable (const tagEntryInfo *const tag)
 	if (tag->placeholder)
 		return false;
 
+	if (! isLanguageEnabled(tag->langType) )
+		return false;
+
 	if (! isLanguageKindEnabled(tag->langType, tag->kindIndex))
 		return false;
 
@@ -1635,7 +1761,7 @@ static void writeTagEntry (const tagEntryInfo *const tag)
 
 	DebugStatement ( debugEntry (tag); )
 
-#ifdef WIN32
+#ifdef _WIN32
 	if (getFilenameSeparator(Option.useSlashAsFilenameSeparator) == FILENAME_SEP_USE_SLASH)
 	{
 		Assert (((const tagEntryInfo *)tag)->inputFileName);
@@ -1700,6 +1826,7 @@ extern void corkTagFile (unsigned int corkFlags)
 		TagFile.corkQueue = ptrArrayNew (deleteTagEnry);
 		tagEntryInfo *nil = newNilTagEntry (corkFlags);
 		ptrArrayAdd (TagFile.corkQueue, nil);
+		TagFile.intervaltab = RB_ROOT;
 	}
 }
 
@@ -1803,8 +1930,9 @@ extern int makeTagEntry (const tagEntryInfo *const tag)
 	if (tag->name [0] == '\0' && (!tag->placeholder))
 	{
 		if (!doesInputLanguageAllowNullTag())
-			error (WARNING, "ignoring null tag in %s(line: %lu)",
-				   getInputFileName (), tag->lineNumber);
+			error (NOTICE, "ignoring null tag in %s(line: %lu, language: %s)",
+				   getInputFileName (), tag->lineNumber,
+				   getLanguageName (tag->langType));
 		goto out;
 	}
 
@@ -1887,9 +2015,8 @@ extern int makeQualifiedTagEntry (const tagEntryInfo *const e)
 extern void setTagPositionFromTag (tagEntryInfo *const dst,
 								   const tagEntryInfo *const src)
 {
-		dst->lineNumber = src->lineNumber;
-		dst->boundaryInfo = src->boundaryInfo;
-		dst->filePosition = src->filePosition;
+	updateTagLine (dst, src->lineNumber, src->filePosition);
+	dst->boundaryInfo = src->boundaryInfo;
 }
 
 static void initTagEntryFull (tagEntryInfo *const e, const char *const name,
@@ -2017,8 +2144,8 @@ static void markTagExtraBitFull (tagEntryInfo *const tag, xtagType extra, bool m
 	else
 	{
 		Assert (extra < countXtags ());
-
-		int n = countXtags () - XTAG_COUNT;
+		Assert (XTAG_COUNT <= countXtags ());
+		unsigned int n = countXtags () - XTAG_COUNT;
 		tag->extraDynamic = xCalloc ((n / 8) + 1, uint8_t);
 		if (!tag->inCorkQueue)
 			PARSER_TRASH_BOX(tag->extraDynamic, eFree);
@@ -2076,6 +2203,42 @@ extern bool isTagExtra (const tagEntryInfo *const tag)
 	return false;
 }
 
+extern void resetTagCorkState (tagEntryInfo *const tag,
+							   enum resetTagMemberAction xtagAction,
+							   enum resetTagMemberAction parserFieldsAction)
+{
+	tagEntryInfo original = *tag;
+
+	tag->inCorkQueue = 0;
+
+	switch (xtagAction)
+	{
+	case RESET_TAG_MEMBER_COPY:
+		copyExtraDynamic (&original, tag);
+		break;
+	case RESET_TAG_MEMBER_CLEAR:
+		tag->extraDynamic = NULL;
+		break;
+	case RESET_TAG_MEMBER_DONTTOUCH:
+		break;
+	}
+
+	switch (parserFieldsAction)
+	{
+	case RESET_TAG_MEMBER_COPY:
+		tag->usedParserFields = 0;
+		tag->parserFieldsDynamic = NULL;
+		copyParserFields (&original, tag);
+		break;
+	case RESET_TAG_MEMBER_CLEAR:
+		tag->usedParserFields = 0;
+		tag->parserFieldsDynamic = NULL;
+		break;
+	case RESET_TAG_MEMBER_DONTTOUCH:
+		break;
+	}
+}
+
 static void assignRoleFull (tagEntryInfo *const e, int roleIndex, bool assign)
 {
 	if (roleIndex == ROLE_DEFINITION_INDEX)
@@ -2162,8 +2325,13 @@ extern void setTagFilePosition (MIOPos *p, bool truncation)
 
 
 	long t0 = 0;
-	if (truncation)
+	if (truncation) {
 		t0 = mio_tell (TagFile.mio);
+		if (t0 == -1)
+			error (FATAL|PERROR,
+				   "failed to tell the file position of the tag file (t0)\n");
+	}
+
 
 	if (mio_setpos (TagFile.mio, p) == -1)
 		error (FATAL|PERROR,
@@ -2172,6 +2340,10 @@ extern void setTagFilePosition (MIOPos *p, bool truncation)
 	if (truncation)
 	{
 		long t1 = mio_tell (TagFile.mio);
+		if (t1 == -1)
+			error (FATAL|PERROR,
+				   "failed to tell the file position of the tag file (t1)\n");
+
 		if (!mio_try_resize (TagFile.mio, (size_t)t1))
 			error (FATAL|PERROR,
 				   "failed to truncate the tag file %ld -> %ld\n", t0, t1);
@@ -2194,3 +2366,47 @@ extern void markAllEntriesInScopeAsPlaceholder (int index)
 {
 	foreachEntriesInScope (index, NULL, markAsPlaceholderRecursively, NULL);
 }
+
+extern int queryIntervalTabByLine(unsigned long lineNum)
+{
+	return queryIntervalTabByRange(lineNum, lineNum);
+}
+
+extern int queryIntervalTabByRange(unsigned long start, unsigned long end)
+{
+	int index = CORK_NIL;
+
+	tagEntryInfoX *ex = intervaltab_iter_first(&TagFile.intervaltab, start, end);
+	while (ex) {
+		index = ex->corkIndex;
+		ex = intervaltab_iter_next(ex, start, end);
+	}
+	return index;
+}
+
+extern int queryIntervalTabByCorkEntry(int corkIndex)
+{
+	Assert (corkIndex != CORK_NIL);
+
+	tagEntryInfoX *ex = ptrArrayItem (TagFile.corkQueue, corkIndex);
+	tagEntryInfo *e = &ex->slot;
+
+	if (e->extensionFields._endLine == 0)
+		return queryIntervalTabByLine(e->lineNumber);
+	return queryIntervalTabByRange(e->lineNumber, e->extensionFields._endLine);
+}
+
+extern bool removeFromIntervalTabMaybe(int corkIndex)
+{
+	if (corkIndex == CORK_NIL)
+		return false;
+
+	tagEntryInfoX *ex = ptrArrayItem (TagFile.corkQueue, corkIndex);
+	tagEntryInfo *e = &ex->slot;
+	if (!e->inIntevalTab)
+		return false;
+
+	intervaltab_remove(ex, &TagFile.intervaltab);
+	e->inIntevalTab = 0;
+	return true;
+}
